require "thrift_client/pool"

class ThreadConnectionPool < ConnectionPool

  def initialize(opts, &block)
    super
    @mutex = Mutex.new
    @condition = ConditionVariable.new
  end

	# Choose first idle connection and pass it to the supplied
  # block. This will block indefinitely until there is an idle
  # connection to service the request.
  def execute
    t = Thread.current
    begin
      conn = acquire(t)
      yield conn
    ensure
      release(t)
    end
  end

  # Destroy all connections in this pool. It will waiting until
  # all connections are idle and than close them one by one.
  def destroy
    while @idle.size < @size
      @mutex.synchronize do
        @condition.wait(@mutex)
      end
    end
    @idle.each do | conn |
      begin
        conn.disconnect
      rescue Exception => e
        puts "close connection error! #{e.backtrace}"
      end
    end
  end

  private

  # Acquire a lock on a connection and assign it to executing thread
  # - if connection is idle, pass it back to the calling block
  # - if pool is full, stop the current thread until connection is idle
  def acquire(thread)
    @mutex.synchronize do
      if conn = @idle.pop
        @active[thread.object_id] = conn
        return conn
      else
        @pending.push thread
        @condition.wait(@mutex)
        @pending.shift
      end
    end
    acquire(thread)
  end

  # Release connection assigned to the supplied thread and
  # wakeup any other pending connections (which will
  # immediately try to run acquire on the pool)
  def release(thread)
    @mutex.synchronize do
      @idle.push(@active.delete(thread.object_id))
      @condition.broadcast
    end
  end
end
