require 'eventmachine'
require 'fiber'

module Thrift
  class TransportException
    ALREADY_CLOSE = 5
  end
  class EventMachineTransport < BaseTransport
    def initialize(host, port=9090, timeout=nil)
      @host, @port, @timeout = host, port, timeout
      @connection = nil
    end

    def open?
      @connection && @connection.connected?
    end

    def open
      @connection = EventMachine.connect(@host, @port, EventMachineConnection) do |conn|
        conn.pending_connect_timeout = @timeout
      end

      fiber = Fiber.current

      @connection.callback {  |arg|
        if fiber == Fiber.current
          return arg
        else
          fiber.resume(arg)
        end
      }
      @connection.errback {  |arg|
        if fiber == Fiber.current
          return arg
        else
          fiber.resume(arg)
        end
      }

      Fiber.yield

      raise TransportException, TransportException::NOT_OPEN, "Unable to connect to #{@host}:#{@port}" unless @connection.connected?
      @connection
    end

    def close
      @connection.close if @connection && @connection.connected?
    end

    def read(sz)
      raise IOError, "read failed, closed stream." unless open?
      @connection.read(sz,@timeout)
    end

    def write(buf)
      raise IOError, "write failed, closed stream." unless open?
      @connection.write(buf,@timeout)
    end
  end

  class EventMachineConnection < EventMachine::Connection
    include EventMachine::Deferrable
    BUFFER_SIZE = 4096 # 4kB

    def post_init
      @connected = false
      @deferrable = nil
      @rbuf = ''
      @index = 0
    end

    def connected?
      @connected
    end

    def connection_completed
      @connected = true
      succeed
    end

    def unbind
      connected = @connected
      @connected = false
      deferrable = @deferrable
      if connected
        if deferrable
          @deferrable = nil
          deferrable.fail(:unbind)
        else
          puts "connection closed by server, but no deferrable was found."
          fail
        end
      else
        puts "connection closed"
        fail
      end
    end

    def on_read_timeout
      trap do
        deferrable = @deferrable
        if deferrable
          @deferrable = nil
          deferrable.fail(:read)
        else
          puts "read timeout, but no deferrable was found."
          fail
        end
      end
    end

    def close
      trap do
        if @connected
          @connected = false
          close_connection(true)
        end
      end
    end

    def write(buf,timeout)
      callback {
        send_data(buf)
      }
    end

    def read(size,timeout)
      timer = nil
      if timeout and timeout > 0
        timer = EventMachine.add_timer(timeout){
          on_read_timeout
        }
      end
      begin
        data = nil
        if can_read?(size)
          data = yank(size)
        else
          @size = size
          @deferrable ||= EventMachine::DefaultDeferrable.new
          data = EventMachine::Synchrony.sync @deferrable
          if data == :unbind
            raise TransportException.new(TransportException::ALREADY_CLOSE, "connection already closed.")
          elsif data == :read
            raise TransportException.new(TransportException::TIMED_OUT, "read timeout")
          elsif data == :readerror
            raise TransportException.new(TransportException::UNKNOWN, "read unknown error")
          end
        end
      ensure
        if timer
          EventMachine.cancel_timer(timer)
        end
      end
      data
    end

    def receive_data(data)
      trap do
        puts data.length
        @rbuf << data
        if @deferrable and can_read?(@size)
          data = yank(@size)
          @size = nil
          deferrable = @deferrable
          @deferrable = nil
          deferrable.succeed(data)
        end
      end
    end

    private

    def can_read?(size)
      @rbuf.size >= @index + size
    end

    def yank(len)
      data = @rbuf.slice(@index, len)
      @index += len
      @index = @rbuf.size if @index > @rbuf.size
      if @index >= BUFFER_SIZE
        @rbuf = @rbuf.slice(@index..-1)
        @index = 0
      end
      data
    end

    def trap
      begin
        yield
      rescue Exception => e
        puts e.message
        puts e.backtrace.join("\n")
        # deferrable = @deferrable
        # if deferrable
        #   @deferrable = nil
        #   deferrable.fail(:readerror)
        # else
        #   puts "read data error, but no deferrable was found."
        #   fail
        # end
      end
    end

  end

end
